@@ -1,5 +1,5 @@
 # ext/associationproxy.py
-# Copyright (C) 2005-2017 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2018 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -13,12 +13,11 @@ transparent proxied access to the endpoint of an association object.
 See the example ``examples/association/proxied_association.py``.
 
 """
-import itertools
 import operator
 import weakref
 from .. import exc, orm, util
 from ..orm import collections, interfaces
-from ..sql import not_, or_
+from ..sql import or_
 from .. import inspect
 
 
@@ -212,7 +211,12 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
         return (self.local_attr, self.remote_attr)
 
     def _get_property(self):
-        return (orm.class_mapper(self.owning_class).
+        owning_class = self.owning_class
+        if owning_class is None:
+            raise exc.InvalidRequestError(
+                "This association proxy has no mapped owning class; "
+                "can't locate a mapped property")
+        return (orm.class_mapper(owning_class).
                 get_property(self.target_collection))
 
     @util.memoized_property
@@ -244,18 +248,38 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
     def _target_is_object(self):
         return getattr(self.target_class, self.value_attr).impl.uses_objects
 
+    def _calc_owner(self, obj, class_):
+        if obj is not None and class_ is None:
+            target_cls = type(obj)
+        elif class_ is not None:
+            target_cls = class_
+        else:
+            return
+
+        # we might be getting invoked for a subclass
+        # that is not mapped yet, in some declarative situations.
+        # save until we are mapped
+        try:
+            insp = inspect(target_cls)
+        except exc.NoInspectionAvailable:
+            # can't find a mapper, don't set owner. if we are a not-yet-mapped
+            # subclass, we can also scan through __mro__ to find a mapped
+            # class, but instead just wait for us to be called again against a
+            # mapped class normally.
+            return
+
+        # note we can get our real .key here too
+        owner = insp.mapper.class_manager._locate_owning_manager(self)
+        if owner is not None:
+            self.owning_class = owner.class_
+        else:
+            # the proxy is attached to a class that is not mapped
+            # (like a mixin), we are mapped, so, it's us.
+            self.owning_class = target_cls
+
     def __get__(self, obj, class_):
         if self.owning_class is None:
-            try:
-                insp = inspect(class_)
-            except exc.NoInspectionAvailable:
-                pass
-            else:
-                if hasattr(insp, 'mapper'):
-                    self.owning_class = insp.mapper.class_
-
-            if self.owning_class is None:
-                self.owning_class = type(obj)
+            self._calc_owner(obj, class_)
 
         if obj is None:
             return self
@@ -278,7 +302,7 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
 
     def __set__(self, obj, values):
         if self.owning_class is None:
-            self.owning_class = type(obj)
+            self._calc_owner(obj, None)
 
         if self.scalar:
             creator = self.creator and self.creator or self.target_class
@@ -295,7 +319,8 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
 
     def __delete__(self, obj):
         if self.owning_class is None:
-            self.owning_class = type(obj)
+            self._calc_owner(obj, None)
+
         delattr(obj, self.key)
 
     def _initialize_scalar_accessors(self):
@@ -308,11 +333,16 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
     def _default_getset(self, collection_class):
         attr = self.value_attr
         _getter = operator.attrgetter(attr)
-        getter = lambda target: _getter(target) if target is not None else None
+
+        def getter(target):
+            return _getter(target) if target is not None else None
+
         if collection_class is dict:
-            setter = lambda o, k, v: setattr(o, attr, v)
+            def setter(o, k, v):
+                setattr(o, attr, v)
         else:
-            setter = lambda o, v: setattr(o, attr, v)
+            def setter(o, v):
+                setattr(o, attr, v)
         return getter, setter
 
     def _new(self, lazy_collection):
@@ -462,6 +492,7 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
         if target_assoc is not None:
             return self._comparator._criterion_exists(
                 target_assoc.contains(obj)
+                if not target_assoc.scalar else target_assoc == obj
             )
         elif self._target_is_object and self.scalar and \
                 not self._value_is_scalar:
@@ -494,7 +525,9 @@ class AssociationProxy(interfaces.InspectionAttrInfo):
             getattr(self.target_class, self.value_attr) != obj)
 
     def __repr__(self):
-        return "AssociationProxy(%r, %r)" % (self.target_collection, self.value_attr)
+        return "AssociationProxy(%r, %r)" % (
+            self.target_collection, self.value_attr)
+
 
 class _lazy_collection(object):
     def __init__(self, obj, target):
@@ -706,7 +739,7 @@ class _AssociationList(_AssociationCollection):
         return list(self) >= other
 
     def __cmp__(self, other):
-        return cmp(list(self), other)
+        return util.cmp(list(self), other)
 
     def __add__(self, iterable):
         try:
@@ -821,7 +854,7 @@ class _AssociationDict(_AssociationCollection):
         return dict(self) >= other
 
     def __cmp__(self, other):
-        return cmp(dict(self), other)
+        return util.cmp(dict(self), other)
 
     def __repr__(self):
         return repr(dict(self.items()))
@@ -921,9 +954,6 @@ class _AssociationSet(_AssociationCollection):
     def _get(self, object):
         return self.getter(object)
 
-    def _set(self, object, value):
-        return self.setter(object, value)
-
     def __len__(self):
         return len(self.col)
 
